from itertools import groupby
import os
import json
from typing import Any
from jinja2 import Environment,FileSystemLoader

from conclave.testresult_info import TestResultInfo

from .scenario import Scenario
from .task import Task
from .scenario_result import ScenarioResult
from .template_base import TemplateBase
from .helpers import datetimeConverter


class Report(TemplateBase):

    def __init__(self) -> None:
        pass

    def generateReport(self, taskReport: Any, testResult: TestResultInfo):
        templateFileName = "report.html"
        env = Environment(loader=FileSystemLoader(os.path.dirname(__file__)))
        jsContent = self.getTemplatePropertyContent('report.min.js')
        template = env.get_template(templateFileName)

        #print(f"task reports: {taskReport}")

        features = []
        for key, group in groupby(sorted(taskReport,key=lambda x:x["feature"]), lambda x: x["feature"]):
            scenarios = []
            featureStatus="success"
            for t in group:
                if t["status"] == 'failed':
                    featureStatus = 'failed'
                task: Task = t["task"]
                scenario: Scenario = task.scenario
                feature: Any = task.feature
                if t['scenario']:
                    scenarios.append({"detail": t['scenario'].scenario, "status": t["status"], "elapsed": t["elapsed"], "error": t["error"]})
                else:
                    scenarios.append({"detail": scenario.gherkinScenario, "status": t["status"], "elapsed": t["elapsed"], "error": t["error"]})
            features.append({"name": key, "status": featureStatus, "scenarios":scenarios, "description": feature['description']})
        
        for feature in features:
            if all(sc['status'] == 'skipped' for sc in feature['scenarios']):
                feature['status'] = 'skipped'
            elif not any(sc['status'] == 'failed' for sc in feature['scenarios']) and any(sc['status'] == 'skipped' for sc in feature['scenarios']):
                feature['status'] = 'incomplete'
            
        output = template.render(
            js=jsContent,
            data=json.dumps(features, default=datetimeConverter),
            testResult=json.dumps(vars(testResult), default=datetimeConverter)
        )

        self.writeTemplateContent("report_output.html",output)


@property
def concurrent_keywords(self):
    return ['*','Concurrently ', '(background) Given ', '(background) When ', '(background) Then ', '(background) And ']

def match_stepline(self,token):
    keywords = (self.dialect.given_keywords +
                self.dialect.when_keywords +
                self.dialect.then_keywords +
                self.dialect.and_keywords +
                self.dialect.concurrent_keywords +
                self.dialect.but_keywords)
    for keyword in (k for k in keywords if token.line.startswith(k)):
        title = token.line.get_rest_trimmed(len(keyword))
        self._set_token_matched(token, 'StepLine', title, keyword)
        return True

    return False



import time
import sys
import os
from typing import Match
sys.path.append(os.path.join(os.path.dirname(__file__),'../'))
from conclave.task_runner import TaskRunner
from conclave.step import Step
from conclave.world import World
import multiprocessing
from conclave.monitor import Monitor
from conclave.task_runner import TaskRunner
from conclave.task_logger import TaskLogger
import requests

@Step(pattern="^call api$")
def callApi(logger: TaskLogger, world: World,match: Match[str]):
    logger.log(f"callApi called")
    x = requests.get('https://httpbin.org/delay/2')
    logger.log(f"response: {x}")

if __name__ == '__main__':
    mon = Monitor()
    print(f"cpu count: {multiprocessing.cpu_count()}")
    #mon.startMonitor()
    tr = TaskRunner(debugMode=True)
    testResult = tr.run(["io_multi_thread.feature"])

    print("\nprogram elapsed time :", testResult.elapsed)

    #mon.stopMonitor()
    #mon.generateReport()

    tr.generateTimeline()
    tr.generateReport()

    if not testResult.success:
        print(f"Test failed")
        os._exit(1)


import multiprocessing
import time
import queue
import threading


class TaskMonitor(threading.Thread):

    def __init__(self):
        super(TaskMonitor, self).__init__()
        self.daemon = True
        self.cancelled = False
        self.signalQueue = multiprocessing.Manager().Queue()
        self.pids = []

    def run(self):
        while not self.cancelled:
            try:
                msg = self.signalQueue.get(block=False)
                #print(f"task pid: {msg}")
                self.pids.append(msg)
            except queue.Empty:
                pass
            time.sleep(1)

    def cancel(self):
        self.cancelled = True

        



import os
from typing import Any
from jinja2 import Environment, FileSystemLoader

from .template_base import TemplateBase

from .task import Task

class DependencyGraph(TemplateBase):

    def __init__(self) -> None:
        pass
    
    def __buildDependencyGraph(self, taskReport: Any, groups: Any):
        graph = ""
        for k,v in groups.items():
            graph += f"subgraph {k}\n"
            for node in v:
                graph += f"{node}\n"
            graph += f"end\n"
        for t in taskReport:
            task: Task = t["task"]
            if task.depends is not None:
                for dp in task.depends:
                    graph += f"{dp}-->{task.id}[{task.name}]\n"
            if task.dependsGroups is not None:
                for dp in task.dependsGroups:
                    graph += f"{dp}-->{task.id}[{task.name}]\n"
            graph += f"{task.id}[{task.name}]\n"
        return graph


    def generateGraph(self, fileName: str, taskReport: Any, groups: Any):
        templateFileName = "dependency.html"
        env = Environment(loader=FileSystemLoader(os.path.dirname(__file__)))
        jsContent = self.getTemplatePropertyContent('mermaid.min.js')
        template = env.get_template(templateFileName)
        graph = self.__buildDependencyGraph(taskReport,groups)

        output = template.render(
            js=jsContent,
            graph=graph
        )

        self.writeTemplateContent(fileName,output)




import os


class TemplateBase:

    def writeTemplateContent(self, fileName: str, content: str):
        with open(fileName, "w", encoding='utf8') as fh:
            fh.write(content)
    
    def getTemplatePropertyContent(self, fileName: str):
        curdir = os.path.dirname(__file__)
        with open(os.path.join(curdir,fileName), "r", encoding='utf8') as fh:
            return fh.read()


from multiprocessing import Process
from itertools import groupby
import multiprocessing
import os
import time
from typing import Any
import psutil
import datetime
import json
import queue
from jinja2 import Environment, FileSystemLoader
from .template_base import TemplateBase
from .helpers import datetimeConverter
from .scenario_result import ScenarioResult


def getProcessByName(processName: str) -> list[Process]:
    procs: list[Process] = []
    for proc in psutil.process_iter():
        if processName in proc.name():
            procs.append(proc)
    return procs


def writeDataPoints(data, fileName):
     
    datetimestr = datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
    with open(fileName, 'a+') as f:
        f.write(f"{datetimestr}> {json.dumps(data, default=datetimeConverter)}\n")

def getCpuUsageByPid(p: Process):
    iowait = None
    cpunum = None
    if hasattr(psutil.Process, "cpu_num"):
        cpunum = p.cpu_num()
    cputimes = p.cpu_times()
    cpus_percent = p.cpu_percent()
    if hasattr(cputimes, "iowait"):
        iowait = cputimes[4]
    
    return { "pid": p.pid, "cpu": cpus_percent, "iowait": iowait, "core": cpunum}


def runMonitor(q: multiprocessing.Queue):
    fileName = f"monitor.txt"
    selfPid = os.getpid()
    print(f"monitor process running: {selfPid}")
    procList = {}
    while True:
        time.sleep(1)
        procs = getProcessByName("python")
        datapoint = []
        for p in procs:
            if p.pid not in procList and p.pid != selfPid:
                procList[p.pid] = p
            if p.pid != selfPid:
                usage = getCpuUsageByPid(procList[p.pid])
                datapoint.append(usage)
        writeDataPoints(datapoint, fileName)
        try:
            msg = q.get(block=False)
            print(f"got termination signal. Exit monitor now")
            break
        except queue.Empty:
            pass
    

class Monitor(TemplateBase):

    def __init__(self) -> None:
        self.signalQueue = multiprocessing.Queue()
        self.monitorProcess = Process(daemon=True,target=runMonitor, args=(self.signalQueue,))

    def startMonitor(self):
        self.monitorProcess.start()
    
    def stopMonitor(self):
        self.signalQueue.put("STOP")
        self.monitorProcess.join()
    
    def __getAllLines(self, fileName: str):
        with open(fileName, "r", encoding='utf8') as fh:
            lines = [line.rstrip() for line in fh]
            return lines
    
    def __getAllPids(self, taskReport: Any):
        pids = []
        for key, group in groupby(sorted(taskReport,key=lambda x:x["feature"]), lambda x: x["feature"]):
            for t in group:
                scenarioResult: ScenarioResult = t["scenario"]
                if scenarioResult:
                    for step in scenarioResult.steps:
                        if step["pid"] is not None:
                            pids.append(step["pid"])
        
        return pids

        
    def generateReport(self, taskReport: Any):
        templateFileName = "monitor.html"
        dataFile = f"monitor.txt"
        env = Environment(loader=FileSystemLoader(os.path.dirname(__file__)))
        cssContent = self.getTemplatePropertyContent('vis-timeline-graph2d.min.css')
        jsContent = self.getTemplatePropertyContent('vis-timeline-graph2d.min.js')
        template = env.get_template(templateFileName)
        allPids = self.__getAllPids(taskReport)
        allPids = list(dict.fromkeys(allPids))
        pids = []

        allItems = []
        dataLines = self.__getAllLines(dataFile)
        for line in dataLines:
            date = line.split("> ")[0]
            item = line.split("> ")[-1]
            parsedItem = json.loads(item)
            for p in parsedItem:
                p["x"] = date
                p["y"] = p["cpu"]
                p["group"] = "cpu time"
                if p["pid"] in allPids:
                    pids.append(p["pid"])
            allItems += parsedItem
            parsedItem = json.loads(item)
        
        pids = list(dict.fromkeys(pids))
        #print(f"all pids: {pids}")

        #print(f"all itens: {allItems}")

        output = template.render(
            pids=pids,
            css=cssContent,
            js=jsContent,
            allItems=json.dumps(allItems)
        )
        self.writeTemplateContent("monitor_output.html",output)
        if os.path.exists(dataFile):
            os.remove(dataFile)


        



import time
import sys
import os
from typing import Match
sys.path.append(os.path.join(os.path.dirname(__file__),'../'))
from conclave.task_runner import TaskRunner
from conclave.step import Step
from conclave.world import World
import multiprocessing
from conclave.monitor import Monitor
from conclave.task_runner import TaskRunner
from conclave.task_logger import TaskLogger

@Step(pattern="^run for long time$")
def longRunningStep(logger: TaskLogger, world: World,match: Match[str]):
    logger.log(f"long running step")
    time.sleep(3)

if __name__ == '__main__':
    print(f"cpu count: {multiprocessing.cpu_count()}")
    my_pid = os.getpid()
    print(f"process pid: {my_pid}")
    tr = TaskRunner(debugMode=True,timeout=10)
    testResult = tr.run(["timeout.feature"])

    print("\nprogram elapsed time :", testResult.elapsed)
    tr.generateTimeline()
    tr.generateReport()

    if not testResult.success:
        print(f"Test failed")
        os._exit(1)
            



from dataclasses import dataclass, field

@dataclass
class TaskRunnerConfig:
    onlyRunScenarioTags: list[str] = field(default_factory=list)
    featureFiles: list[str] = field(default_factory=list)



from typing import Any
from dataclasses import dataclass

@dataclass
class ScenarioResult:
    scenario: Any = None
    steps: Any = None
    id: str = None
    message: str = None
    elapsed: float = 0.0
    exception: str = None
    threadId: int = None
    pid: int = None
    startTime: Any = None
    endTime: Any = None


import time
import sys
import os
from typing import Match
sys.path.append(os.path.join(os.path.dirname(__file__),'../'))
from conclave.task_runner import TaskRunner
from conclave.step import Step
from conclave.world import World
from conclave.task_runner import TaskRunner
from conclave.task_logger import TaskLogger

@Step(pattern="^step 1$")
def func1(logger: TaskLogger, world: World,match: Match[str]):
    logger.log(f"func1 called")

if __name__ == '__main__':
    my_pid = os.getpid()
    print(f"process pid: {my_pid}")
    tr = TaskRunner(debugMode=True)
    testResult = tr.run(["dep.feature"])

    print("\nprogram elapsed time :", testResult.elapsed)

    tr.generateTimeline()
    tr.generateDependencyGraph()
    tr.generateReport()

    if not testResult.success:
        print(f"Test failed")
        os._exit(1)


from dataclasses import dataclass, field
from typing import Any
from .scenario import Scenario

@dataclass
class Task:
    name: str
    scenario: Scenario
    feature: Any
    id: str
    depends: list[str] = field(default_factory=list)
    dependsGroups: list[str] = field(default_factory=list)
    runAlways: bool = False
    group: str = None
    isSetup: bool = False
    isConcurrent: bool = False
    isTeardown: bool = False
    isParallel: bool = False



import time
import sys
import os
from typing import Match
sys.path.append(os.path.join(os.path.dirname(__file__),'../'))
from conclave.task_runner import TaskRunner
from conclave.step import Step
from conclave.world import World
import multiprocessing
from conclave.monitor import Monitor
from conclave.task_runner import TaskRunner
from conclave.task_logger import TaskLogger

@Step(pattern="^step 1$")
def step1(logger: TaskLogger, world: World,match: Match[str]):
    logger.log(f"step 1")
    world.setProp("itemA","hello")
    time.sleep(10)

@Step(pattern="^step 2$")
def step2(logger: TaskLogger, world: World,match: Match[str]):
    logger.log(f"step 1")
    value = world.getProp("itemA")
    logger.log(f"itemA: {value}")

if __name__ == '__main__':
    print(f"cpu count: {multiprocessing.cpu_count()}")
    tr = TaskRunner(debugMode=True)
    testResult = tr.run(["parallel_world.feature"])

    print("\nprogram elapsed time :", testResult.elapsed)

    if not testResult.success:
        print(f"Test failed")
        os._exit(1)


import time
import sys
import os
from typing import Match
sys.path.append(os.path.join(os.path.dirname(__file__),'../'))
from conclave.task_runner import TaskRunner
from conclave.step import Step
from conclave.world import World
import multiprocessing
from conclave.monitor import Monitor
from conclave.task_runner import TaskRunner
from conclave.task_logger import TaskLogger
from conclave.report import Report

@Step(pattern="^calculate pi$")
def calcPi(logger: TaskLogger, world: World,match: Match[str]):
    logger.log(f"calcPi called")
    pi = 0
    accuracy = 100000

    for i in range(50):
        for i in range(0, accuracy):
            pi += ((4.0 * (-1)**i) / (2*i + 1))
    logger.log(f"done calc pi")

if __name__ == '__main__':
    mon = Monitor()
    print(f"cpu count: {multiprocessing.cpu_count()}")
    my_pid = os.getpid()
    #mon.startMonitor()
    print(f"process pid: {my_pid}")
    tr = TaskRunner(debugMode=True)
    testResult = tr.run(["cpu_multi_thread.feature"])

    print("\nprogram elapsed time :", testResult.elapsed)

    #mon.stopMonitor()
    #mon.generateReport()

    tr.generateTimeline()
    tr.generateReport()

    if not testResult.success:
        print(f"Test failed")
        os._exit(1)


from dataclasses import dataclass
from typing import Any, Callable, Optional
import time
import traceback
import functools
import re
import os
import threading
from datetime import datetime

from .task_logger import TaskLogger
from .step_definition import StepDefinition

@dataclass
class StepResult:
    elapsed: float
    result: Any
    error: str
    threadId: int
    pid: int
    start: Any
    end: Any
    log: str

class Step:

    """
    A decorator class that can be used to mark methods as being a step definition in gherkin
    Example usage:

    @Step(pattern="I login to webiste")
    def mystep(logger, world, match):
        logger.log(f"mystep called")
    """ 
    stepDefinitions: list[StepDefinition] = []

    def __init__(self, pattern: str) -> None:
        self.pattern = pattern

    def __call__(self, func) -> Any:
        
        @functools.wraps(func)
        def wrapper_func(*args,**kwargs):
            start = time.time()
            pid = os.getpid()
            threadId = threading.get_ident()
            result,exc,elapsed = None,None,0.0
            logger = TaskLogger(f"{args[0].funcName}/{func.__name__}")
            parentLogger = args[0]
            newargs = list(args)
            newargs[0] = logger
            args2 = tuple(newargs)
            try:
                result = func(*args2,**kwargs)
            except Exception:
                exc = traceback.format_exc()
            finally:
                end = time.time()
                elapsed = end-start
                parentLogger.msg += logger.msg
                return StepResult(elapsed,result,exc,threadId,pid,datetime.fromtimestamp(start),datetime.fromtimestamp(end),logger.msg)
        Step.stepDefinitions.append(StepDefinition(self.pattern,wrapper_func))
        return wrapper_func
    
    @staticmethod
    def getStep(text: str) -> Optional[Callable]:
        for step in Step.stepDefinitions:
            result = re.search(step.pattern,text)
            if result:
                return step.func,result
        return None,None



import datetime
import os
import glob
from typing import Any, Optional, Union
from gherkin.token_scanner import TokenScanner
from gherkin.parser import Parser

from conclave.report import Report

from .testresult_info import TestResultInfo

from .timeline import Timeline
from .task import Task
import threading
import concurrent
from .color import bcolors
from concurrent.futures import ThreadPoolExecutor, wait, ProcessPoolExecutor
from .scenario import Scenario
from itertools import groupby
import multiprocessing
from gherkin.token_matcher import TokenMatcher
from gherkin.dialect import Dialect
from .custom_keywords import concurrent_keywords, match_stepline
from .dependency_graph import DependencyGraph
import uuid
import time
import signal
from .task_monitor import TaskMonitor
from .task_runner_config import TaskRunnerConfig

Dialect.concurrent_keywords = concurrent_keywords
TokenMatcher.match_StepLine = match_stepline

class TaskRunner:

    def __init__(self,debugMode=False,timeout=3600) -> None:
        self.parser = Parser()
        self.completedTasks: list[str] = []
        self.groups = {}
        self.pool = ThreadPoolExecutor()
        self.parallelPool = ProcessPoolExecutor(max_workers=multiprocessing.cpu_count(),mp_context=multiprocessing.get_context("spawn"))
        self.taskReport = []
        self.setupTasks: list[Task] = []
        self.teardownTasks: list[Task] = []
        self.allTaskIds: list[str] = []
        self.mainTasks: list[Task] = []
        self.debugMode: bool = debugMode
        self.testResult: TestResultInfo = None
        self.timeout = timeout
        self.taskMonitor = TaskMonitor()

    def run(self, options: Union[list[str],TaskRunnerConfig]) -> TestResultInfo:
        
        if isinstance(options,TaskRunnerConfig):
            if len(options.featureFiles) > 0:
                files = self.__getAllFeatureFiles(options.featureFiles)
                for file in files:
                    self.__parse(file, options.onlyRunScenarioTags)
        else:
            files = self.__getAllFeatureFiles(options)
            for file in files:
                self.__parse(file, [])
        
        start = time.time()
        startDate = datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

        self.taskMonitor.start()

        ## run any setup tasks
        error = self.__runSetupTasks()

        if not error:
            ## run main tasks
            self.__runMainTasks()

        ## run any teardown tasks
        error = self.__runTeardownTasks()

        end = time.time()
        endDate = datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

        self.taskMonitor.cancel()
        self.taskMonitor.join()
        self.taskMonitor.pids = list(dict.fromkeys(self.taskMonitor.pids))
        for pid in self.taskMonitor.pids:
            self.__print(f"kill process id: {pid}")
            try:
                os.kill(pid, signal.SIGTERM)
            except:
                pass

        ## print test report
        self.__printTestReport()

        self.testResult = TestResultInfo(
            elapsed=end-start,
            start=startDate,
            end=endDate,
            numCpu=multiprocessing.cpu_count(),
            pid=os.getpid(),
            success=len(list(filter(lambda x: "failed" in x["status"], self.taskReport))) <= 0
        )

        return self.testResult

    def __getAllFeatureFiles(self, paths: list[str]) -> list[str]:
        featureFiles: list[str] = []
        for p in paths:
            fullPath = os.path.abspath(p)
            if os.path.isfile(fullPath):
                if fullPath not in featureFiles:
                    featureFiles.append(fullPath)
            elif os.path.isdir(fullPath):
                files = glob.glob(fullPath + '/**/*.feature', recursive=True)
                for file in files:
                    if file not in featureFiles:
                        featureFiles.append(file)
        return featureFiles
                

    def __parse(self, featureFile: str, onlyTags: list[str]):
        result = self.parser.parse(TokenScanner(featureFile))
        if not result:
            return
        
        if "feature" not in result:
            return

        if "children" not in result["feature"]:
            return
        
        feature = result["feature"]
        for child in feature["children"]:
            if "scenario" in child:
                sc = child["scenario"]
                self.__getScenario(sc, feature, onlyTags)
    
    # def __getFeatureId(self, feature: Any) -> str:
    #     tags = feature["tags"]
    #     id=uuid.uuid4().hex()
    #     for tag in tags:
    #         tg = tag["name"]
    #         if temp := self.__getIdTag(tg, tag): id = temp
    #     return id
    
    def __getScenario(self, scenario: Any, feature: Any, onlyIncludeTags: list[str]) -> Optional[Task]:
        tags = scenario["tags"]
        isConcurrent,id,depends,dependsGroups,group,runAlways,isSetup,isTeardown,isParallel = False,uuid.uuid4().hex,[],[],None,False,False,False,False
        
        # if scenario tags filtering is specified an scenario does not have any tag then exclude this scenario
        if len(onlyIncludeTags) > 0 and len(tags) <= 0:
            return None

        # filter scenario based on tags
        if len(onlyIncludeTags) > 0:
            if not any(item["name"] in onlyIncludeTags for item in tags):
                return None

        for tag in tags:
            tg = tag["name"]
            if temp := self.__getConcurrentTag(tg): isConcurrent = temp
            if temp := self.__getParallelTag(tg): isParallel = temp
            if temp := self.__getRunAlwaysTag(tg): runAlways = temp
            if temp := self.__getSetupTag(tg): isSetup = temp
            if temp := self.__getTeardownTag(tg): isTeardown = temp
            if temp := self.__getIdTag(tg, tag): id = temp
            if temp := self.__getDependsTag(tg, tag): depends.append(temp)
            if temp := self.__getDependsGroupsTag(tg, tag): dependsGroups.append(temp)
            if temp := self.__getGroupTag(tg, tag): group = temp
        sc = Scenario(scenario["name"],scenario,feature,id)
        t = Task(scenario["name"], sc, feature, id,depends,dependsGroups,runAlways,group, isSetup, isConcurrent,isTeardown,isParallel)
        self.allTaskIds.append(t.id)
        if group is not None:
            self.groups.setdefault(group, []).append(id)
        if isSetup:
            self.setupTasks.append(t)
        elif isTeardown:
            self.teardownTasks.append(t)
        else:
            self.mainTasks.append(t)
        return t

    def __getGroupTag(self, name: str, tag: Any) -> Optional[str]:
        return tag["name"].split("@group_")[-1] if name.startswith("@group_") else None

    def __getIdTag(self, name: str, tag: Any) -> Optional[str]:
        return tag["name"].split("@id_")[-1] if name.startswith("@id_") else None
    
    def __getConcurrentTag(self, name: str) -> bool:
        return name == "@concurrent"
    
    def __getParallelTag(self, name: str) -> bool:
        return name == "@parallel"
    
    def __getRunAlwaysTag(self, name: str) -> bool:
        return name == "@runAlways"

    def __getDependsTag(self, name: str, tag: Any) -> Optional[str]:
        return tag["name"].split("@depends_")[-1] if name.startswith("@depends_") else None
    
    def __getDependsGroupsTag(self, name: str, tag: Any) -> Optional[str]:
        return tag["name"].split("@dependsGroups_")[-1] if name.startswith("@dependsGroups_") else None
    
    def __getSetupTag(self, name: str) -> bool:
        return name == "@setup"
    
    def __getTeardownTag(self, name: str) -> bool:
        return name == "@teardown"

    def __getNextTask(self,taskList) -> list[Task]:
        new_tasks: list[Task] = []
        skipped_tasks: list[Task] = []
        self.__print(f"tasks pending: {[(f'name: {t.name}',f'id:{t.id}') for t in taskList]}")
        for task in taskList:
            self.__print(f"check pending task: (name:{task.name},id:{task.id})")
            if len(task.depends) > 0:
                self.__print(f"task (name:{task.name},id:{task.id}) depends on: {task.depends}")
                if not set(task.depends).issubset(self.allTaskIds):
                    self.__addTaskToReport(task, "skipped", None, 0.0, None)
                    skipped_tasks.append(task)
                    self.completedTasks.append(task.id)
                    continue
                if not set(task.depends).issubset(self.completedTasks):
                    continue
                if self.__isParentTaskFailed(task.depends) and not task.runAlways:
                    self.__addTaskToReport(task, "skipped", None, 0.0, None)
                    self.__print(f"skip task: (name:{task.name},id:{task.id})")
                    skipped_tasks.append(task)
                    self.completedTasks.append(task.id)
                    continue
            if len(task.dependsGroups) > 0:
                self.__print(f"task (name:{task.name},id:{task.id}) depends on groups: {task.dependsGroups}")
                combine_groups = []
                for g in task.dependsGroups:
                    if g in self.groups:
                        combine_groups += self.groups[g]
                if not bool(combine_groups):
                    self.__print(f"no groups matching found for task: (name:{task.name},id:{task.id})")
                    self.__addTaskToReport(task, "skipped", None, 0.0, None)
                    skipped_tasks.append(task)
                    self.completedTasks.append(task.id)
                    continue
                if not set(combine_groups).issubset(self.completedTasks):
                    continue
                if self.__isParentTaskFailed(combine_groups) and not task.runAlways:
                    self.__addTaskToReport(task, "skipped", None, 0.0, None)
                    self.__print(f"dependent tasks failed so skip task: (name:{task.name},id:{task.id})")
                    skipped_tasks.append(task)
                    self.completedTasks.append(task.id)
                    continue
            new_tasks.append(task)
        if len(skipped_tasks) > 0:
            taskList = [i for i in taskList if not any(x.id == i.id for x in skipped_tasks)]
        if len(new_tasks) > 0:
            taskList = [i for i in taskList if not any(x.id == i.id for x in new_tasks)]
        return new_tasks, taskList

    def runWorkerThread(self, taskList):
        tasks_to_submit, taskList = self.__getNextTask(taskList)
        futures = {}
        for task in tasks_to_submit:
            if task.isConcurrent:
                futures[self.pool.submit(task.scenario.run,queue=None)] = task
            elif task.isParallel:
                futures[self.parallelPool.submit(task.scenario.run,queue=self.taskMonitor.signalQueue)] = task
        
        
        self.__print(f"adding new tasks: {[(f'name: {t.name}',f'id:{t.id}') for t in tasks_to_submit]}")
        self.__print(f"tasks in pool: {[(f'name: {t.name}',f'id:{t.id}') for t in futures.values()]}")

        startTime = time.time()
        currentTimeout = self.timeout
        while futures:
            done, notDone = wait(futures,return_when=concurrent.futures.FIRST_COMPLETED,timeout=currentTimeout)
            endTime = time.time()
            elapsed = endTime-startTime
            self.__print(f"elapsed time waiting for task to complete: {elapsed}")
            if elapsed >= currentTimeout:
                for c in notDone:
                    self.__print(f"tasks not done: (name:{futures[c].name},id:{futures[c].id}, running:{c.running()},cancelled:{c.cancelled()})")
                    self.__addTaskToReport(futures[c],"failed","timeout waiting for task to complete",self.timeout, None)
                self.__print(f"timeout waiting {self.timeout} (s) for remaining tasks to complete. Aborting.")
                break
            currentTimeout = currentTimeout - elapsed
            self.__print(f"Remaining timeout: {currentTimeout}")
            for c in done:
                fut = futures.pop(c)
                result = c.result()
                print(result.message)
                self.__print(f"task completed: (name:{fut.name},id:{result.id})")
                self.completedTasks.append(result.id)
                if result.exception is not None:
                    self.__addTaskToReport(fut,"failed",result.exception,result.elapsed, result)
                else:
                    self.__addTaskToReport(fut,"success",result.exception,result.elapsed, result)
                next_tasks,taskList = self.__getNextTask(taskList)
                if next_tasks is not None:
                    for t in next_tasks:
                        self.__print(f"adding new task (name:{t.name},id:{t.id})")
                        if t.isConcurrent:
                            item = self.pool.submit(t.scenario.run,queue=None)
                            futures[item] = t
                        elif t.isParallel:
                            item = self.parallelPool.submit(t.scenario.run,queue=self.taskMonitor.signalQueue)
                            futures[item] = t
                self.__print(f"remaining tasks in pool: {[(f'name: {t.name}',f'id:{t.id}') for t in futures.values()]}")
    
    def __printTestReport(self):
        print(f"Test report:\n")

        for key, group in groupby(sorted(self.taskReport,key=lambda x:x["feature"]), lambda x: x["feature"]):
            print(f"\nFeature: {key}\n")
            for t in group:
                if t['status'] == 'success':
                    print(f"\tScenario: {t['name']}: {bcolors.OKGREEN}{t['status']}{bcolors.ENDC} (elapsed {t['elapsed']})")
                elif t['status'] == 'skipped':
                    print(f"\tScenario: {t['name']}: {bcolors.WARNING}{t['status']}{bcolors.ENDC} (elapsed {t['elapsed']})")
                else:
                    print(f"\tScenario: {t['name']}: {bcolors.FAIL}{t['status']}{bcolors.ENDC} (elapsed {t['elapsed']})")
                
                # if t['error'] is not None:
                #     print(f"\t\tError: {bcolors.FAIL}{t['error']}{bcolors.ENDC}")
                
                #print(f"all steps report: {t['scenario']}")
                if 'scenario' in t and t["scenario"] is not None:
                    if t['scenario'].steps:
                        for step in t["scenario"].steps:
                            status = step['status']
                            if status == 'failed':
                                print(f"\t  Step: {step['keyword']}{step['text']}{bcolors.FAIL} {step['status']}{bcolors.ENDC} (elapsed {step['elapsed']})")
                            elif status == 'skipped':
                                print(f"\t  Step: {step['keyword']}{step['text']}{bcolors.WARNING} {step['status']}{bcolors.ENDC} (elapsed {step['elapsed']})")
                            else:
                                print(f"\t  Step: {step['keyword']}{step['text']}{bcolors.OKGREEN} {step['status']}{bcolors.ENDC} (elapsed {step['elapsed']})")
                            if "error" in step and step['error'] is not None:
                                print(f"\t\t{bcolors.FAIL}{step['error']}{bcolors.ENDC}")

                    
    
    def __print(self,msg: str):
        if self.debugMode:
            print(f"{bcolors.OKCYAN}[{datetime.datetime.now().strftime('%m/%d/%Y, %H:%M:%S')} task_manager] {msg}{bcolors.ENDC}\n")
    

    def __isParentTaskFailed(self, groups):
        ptask = []
        for t in self.taskReport:
            if 'id' in t:
                if any(t['id'] == x for x in groups):
                    ptask.append(t)
        return any(y['status'] == "failed" or y['status'] == "skipped" for y in ptask)
    
    def __addTaskToReport(self, task: Task, status: str, error: str, elapsed: float, scenarioResult: Any):
        if not any(task.name == x["name"] and task.feature["name"] == x["feature"] for x in self.taskReport):
            self.taskReport.append({"name":task.name,"status":status,"error":error, "elapsed": elapsed, "id": task.id, "feature": task.feature["name"], "task": task, "scenario": scenarioResult})

    def __runMainTasks(self):
        return self.__runTasks(self.mainTasks)

    def __runSetupTasks(self):
        return self.__runTasks(self.setupTasks)
    
    def __runTeardownTasks(self):
        return self.__runTasks(self.teardownTasks)
    
    def __transformSeqTasks(self, taskList: list[Task]):
        dependTaskId = None
        for t in taskList:
            t.isConcurrent = True
            if dependTaskId is not None and len(t.depends) <= 0 and len(t.dependsGroups) <= 0:
                t.depends.append(dependTaskId)
            dependTaskId = t.id
        return taskList

    def __runTasks(self, taskList):
        error = False
        seqtasks = list(filter(lambda x: not x.isConcurrent and not x.isParallel, taskList))
        contasks = list(filter(lambda x: x.isConcurrent or x.isParallel, taskList))
        seqtasks = self.__transformSeqTasks(seqtasks)
        alltasks = contasks + seqtasks
        self.__print(f"all tasks: {[t.name for t in alltasks]}")

        workerThread = threading.Thread(target=self.runWorkerThread,kwargs={'taskList':alltasks})
        workerThread.start()
        
        workerThread.join()

        if len(list(filter(lambda x: "failed" in x["status"], self.taskReport))) > 0:
            error = True
        
        return error
    
    def generateTimeline(self):
        timeline = Timeline()
        timeline.generateTimeline(self.taskReport)
    
    def generateDependencyGraph(self):
        depGraph = DependencyGraph()
        depGraph.generateGraph("dependency_output.html",self.taskReport,self.groups)
    
    def generateReport(self):
        report = Report()
        report.generateReport(self.taskReport, self.testResult)
    




from .color import bcolors
import datetime

class TaskLogger:
    def __init__(self, funcName):
        self.msg = ""
        self.funcName = funcName

    def log(self,msg):
        self.msg += f"[{datetime.datetime.now().strftime('%m/%d/%Y, %H:%M:%S')} {self.funcName}] {msg}\n"
    def error(self,msg):
        self.msg += f"{bcolors.FAIL}[{datetime.datetime.now().strftime('%m/%d/%Y, %H:%M:%S')} {self.funcName}] {msg}{bcolors.ENDC}\n"


from distutils.core import setup

setup(name="paraworld",
      packages=["conclave"],
      version="1.1.1",
      description="A BDD framework for running parallel and concurrent tests",
      long_description="A BDD framework for writing concurrent and parallel tests",
      author="DasAng",
      author_email="",
      url="https://github.com/DasAng/paraworld",
      license="GPL",
      download_url="http://pypi.python.org/pypi/paraworld",
      keywords=["gherkin", "parallel", "bdd"],
      scripts=[],
      classifiers=["Programming Language :: Python",
                   "Programming Language :: Python :: 3",
                   ],
      platforms = ['any'],
      install_requires=['gherkin-official==22.0.0', 'jinja2==3.0.2','psutil==5.8.0','requests==2.26.0'],
      include_package_data=True
      )


from threading import Lock
from typing import Any
import multiprocessing

class Singleton(type):
    __instances = {}
    __lock: Lock = Lock()

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        with self.__lock:
            if self not in self.__instances:
                instance = super().__call__(*args, **kwds)
                self.__instances[self] = instance
        return self.__instances[self]


class World(metaclass=Singleton):

    def __init__(self) -> None:
        self.__props = multiprocessing.Manager().dict()

    def setProp(self,key: Any, value: Any):
        self.__props[key] = value
    
    def getProp(self,key: Any) -> Any:
        if key in self.__props:
            return self.__props[key]
    
    def __getstate__(self):
        state = self.__dict__.copy()
        return state
    
    def __setstate__(self,state):
        self.__dict__.update(state)

    


from concurrent.futures import ThreadPoolExecutor, wait
import concurrent
import multiprocessing
from typing import Any
from datetime import datetime
from .scenario_result import ScenarioResult

from .task_logger import TaskLogger
import traceback
import time
import os
import threading
from .step import Step
from .world import World

class Scenario:

    """
    This class represents a scenario in Gherkin. A scenario consists of steps which
    will be executed one by one. All background steps will be executed before the
    steps of a scenario
    """
    def __init__(self, name, gherkinScenario,gherkinFeature,id):
        self.name = name
        self.logger = TaskLogger(name)
        self.gherkinScenario = gherkinScenario
        self.id = id
        self.gherkinFeature = gherkinFeature
        self.steps = self.__getAllSteps()
        self.world = World()
        self.pool = ThreadPoolExecutor()
        self.workerQueue = []
        self.lock = threading.Lock()
        self.stopWorker = False
        self.stepsError = {}
        self.result = ScenarioResult(scenario=self.gherkinScenario,id=self.id,steps=self.steps,threadId=None,pid=None,startTime=None,endTime=None)

    def run(self, queue: multiprocessing.Queue):
        """
        Execute the scenario
        """
        self.logger.log(f"Run scenario: {self.name}")
        my_pid = os.getpid()
        if queue:
            queue.put(my_pid)
        self.logger.log(f"process pid: {my_pid}")
        self.logger.log(f"thread id: {threading.get_ident()}")
        start = time.time()
        exc = None
        try:
            self.__runSteps()
            if self.stepsError:
                for key in self.stepsError:
                    exc = f"Concurrent/parallel step: {key}\n{self.stepsError[key]}\n"
        except Exception:
            exc = f"failed: {traceback.format_exc()}"
            if self.stepsError:
                for key in self.stepsError:
                    exc += f"\nConcurrent/parallel step: {key}\n{self.stepsError[key]}\n"
            self.logger.error(f"{exc}")
        finally:
            end = time.time()
            elapsed = end-start
            self.logger.log(f"elapsed time: {end-start}")
            message = self.logger.msg
            self.result.message = message
            self.result.exception = exc
            self.result.elapsed = elapsed
            self.result.threadId = threading.get_ident()
            self.result.pid = os.getpid()
            self.result.startTime = datetime.fromtimestamp(start)
            self.result.endTime = datetime.fromtimestamp(end)
            return self.result

    
    def msg(self):
        return self.logger.msg
    

    def __getBackgroundSteps(self):
        """
        Get all background steps for a scenario
        """
        backgroundSteps = []
        for child in self.gherkinFeature["children"]:
            if "background" in child:
                bg = child["background"]
                if "steps" in bg:
                    backgroundSteps.extend(bg["steps"])
        return backgroundSteps

    def __getSteps(self):
        """
        Get all main steps for a scenario
        """
        steps = []
        if "steps" in self.gherkinScenario:
            steps.extend(self.gherkinScenario["steps"])
        return steps

    def __getAllSteps(self):
        """
        Get all steps for a scenario. This will be a combined list of background steps and main steps
        """
        bgSteps = self.__getBackgroundSteps()
        steps = self.__getSteps()
        bgSteps.extend(steps)
        return bgSteps

    def __runSteps(self):
        """
        Execute all steps in the scenario
        """
        workerThread = None

        try:
            allSteps = self.steps

            for step in self.steps:
                self.__updateStep(step, "skipped", None, 0.0, None, None,None,None, "")

            workerThread = threading.Thread(target=self.runWorkerThread,kwargs={'taskList':[]})
            workerThread.start()
                

            for step in allSteps:
                exc = None
                self.logger.log(f"execute step: {step['keyword']} {step['text']}")
                func,match = Step.getStep(step['text'])
                if func:
                    if self.__isStepConcurrent(step):
                        self.logger.log(f"add step to worker queue: {step['keyword']} {step['text']}")
                        self.__addStepToWorkerPool(step)
                    else:
                        result = func(self.logger,self.world,match)
                        if result.error:
                            self.__updateStep(step, "failed", result.error, result.elapsed, result.pid,result.threadId,result.start,result.end,result.log)
                            raise Exception(result.error)
                        else:
                            self.__updateStep(step, "success", result.error, result.elapsed,result.pid,result.threadId,result.start,result.end,result.log)
                else:
                    self.__updateStep(step, "skipped", f"Could not find matching step definition for: {step['keyword']}{step['text']}",0.0,None,None,None,None,"")
                    raise Exception(f"Could not find matching step definition for: {step['keyword']}{step['text']}")
                    
        finally:
            self.stopWorker = True
            if workerThread is not None:
                workerThread.join()
    
    def __updateStep(self, step: Any, status: str,error: str, elapsed: int, pid: int,threadId: int,start: Any, end: Any, log: str):
        item = next((x for x in self.steps if x['id'] == step['id']), None)
        if item:
            item["status"] = status
            step["error"] = error
            step["elapsed"] = elapsed
            step["pid"] = pid
            step["threadId"] = threadId
            step["start"] = start
            step["end"] = end
            step["log"] = log

    def __addStepToWorkerPool(self, step):
        with self.lock:
            self.workerQueue.append(step)

    def __isStepConcurrent(self, step):
        return step['keyword'] in ('Concurrently ', '(background) Given ','(background) When ','(background) Then ','(background) And ')
        # return step['keyword'] == 'Concurrently ' or \
        # step['keyword'] == '(background) Given ' or \
        # step['keyword'] == '(background) Then ' or \
        # step['keyword'] == '(background) When ' or \
        # step['keyword'] == '(background) And '


    def __getAllConcurrentSteps(self):
        """
        Retrieve all steps marked to be run concurrently
        """
        return list(filter(lambda x: x["keyword"] == "Concurrently ", self.steps))
    
    def __getAllSequentialSteps(self):
        """
        Retrieve all sequential steps
        """
        return list(filter(lambda x: x["keyword"] != "Concurrently ", self.steps))
    
    def __getNextSteps(self):
        with self.lock:
            steps, self.workerQueue = self.workerQueue, []
        return steps

    def runWorkerThread(self, taskList):
        futures = {}
        for step in taskList:
            func,match = Step.getStep(step['text'])
            if func:
                futures[self.pool.submit(func, self.logger, self.world,match)] = step
            else:
                self.__updateStep(step, "skipped", f"Could not find matching step definition for: {step['keyword']}{step['text']}",0.0,None,None,None,None,"")
        
        self.logger.log(f"add steps for concurrent executions: {[t['keyword']+t['text'] for t in futures.values()]}")

        self.stepsError = {}
        while True:
            # fix issue with 100% CPU usage when there are no concurrent tasks
            # but we still call wait() with timeout of 1 seconds. This causes
            # a busy wait loop that eats up all CPU time. Instead we will check
            # if we have any concurrent tasks and then just sleep for 1 seconds
            # before the next check
            if not futures:
                time.sleep(1)
            else:
                done, _ = wait(futures,return_when=concurrent.futures.FIRST_COMPLETED,timeout=1)
                for c in done:
                    fut = futures.pop(c)
                    result = c.result()
                    if result.error:
                        self.stepsError[fut["keyword"]+fut["text"]] = result.error
                        self.logger.error(f"failed: {result.error}")
                        self.__updateStep(fut,"failed",result.error,result.elapsed,result.pid,result.threadId,result.start,result.end,result.log)
                    else:
                        self.__updateStep(fut,"success",None,result.elapsed,result.pid,result.threadId,result.start,result.end,result.log)
                    
                    self.logger.log(f"completed step: {fut['keyword']} {fut['text']}")
            nextSteps = self.__getNextSteps()
            for step in nextSteps:
                func,match = Step.getStep(step['text'])
                if func:
                    self.logger.log(f"add concurrent step for execution: {step['text']}")
                    futures[self.pool.submit(func, self.logger, self.world,match)] = step
                else:
                    self.__updateStep(step, "skipped", f"Could not find matching step definition for: {step['keyword']}{step['text']}",0.0,None,None,None,None,"")
            if not futures and self.stopWorker:
                # To avoid scenarios where we have only concurrent steps remaining in the queue and the scenario has no more steps
                # to execute. Then the main thread will set the variable stopWorker to True, but this thread has not gotten a chance
                # to retrieve the pending concurrent steps to be executed. So we try and fetch any pending items one last time before
                # exiting if no pending steps are found.
                nextSteps = self.__getNextSteps()
                for step in nextSteps:
                    func,match = Step.getStep(step['text'])
                    if func:
                        self.logger.log(f"add concurrent step for execution: {step['text']}")
                        futures[self.pool.submit(func, self.logger, self.world,match)] = step
                    else:
                        self.__updateStep(step, "skipped", f"Could not find matching step definition for: {step['keyword']}{step['text']}",0.0,None,None,None,None,"")
                if not futures:
                    break
    
    def __getstate__(self):
        state = self.__dict__.copy()
        del state["pool"]
        del state["lock"]
        return state
    
    def __setstate__(self,state):
        self.__dict__.update(state)
        self.pool = ThreadPoolExecutor()
        self.lock = threading.Lock()
             



from itertools import groupby
from typing import Any
import json
import os
import uuid
import datetime
from jinja2 import Environment, FileSystemLoader
from .scenario_result import ScenarioResult
from .template_base import TemplateBase
from .helpers import datetimeConverter


class Timeline(TemplateBase):

    def __init__(self) -> None:
        self.templateEnv = Environment(loader=FileSystemLoader(os.path.dirname(__file__)))

    def __getAllPids(self, taskReport: Any):
        processList = {}
        pgroups = []
        threadIds = []
        pitems = []
        for key, group in groupby(sorted(taskReport,key=lambda x:x["feature"]), lambda x: x["feature"]):
            for t in group:
                scenarioResult: ScenarioResult = t["scenario"]
                if scenarioResult:
                    for step in scenarioResult.steps:
                        if step["pid"] is not None:
                            processList.setdefault(step["pid"], []).append(step)
        
        #print(f"process list: {processList}")
        for key,val in processList.items():
            threadIds = [x["threadId"] for x in val]
            threadIds = list(dict.fromkeys(threadIds))
            for threadId in threadIds:
                pgroups.append({
                    "id": f"{key}_{threadId}",
                    "content": f"thread {threadId}"
                })
            threadIdGroups = [f"{key}_{x}" for x in threadIds]
            pgroups.append({
                "id": key,
                "content": f"Process Id {key}",
                "nestedGroups": threadIdGroups,
                "treeLevel": 1
            })
            for item in val:
                if item["threadId"] is not None:
                    pitems.append({
                        "id": uuid.uuid4().hex,
                        "content": "",
                        "group": f"{item['pid']}_{item['threadId']}",
                        "start": item["start"].strftime("%m/%d/%Y, %H:%M:%S"),
                        "end": item["end"].strftime("%m/%d/%Y, %H:%M:%S"),
                        "type": "background"
                    })
        #print(f"all thread ids: {threadIds}")
        return pgroups,pitems

    def __createFeatureItem(self,id):
        item = {
            "id": id,
            "content": id,
            "nestedGroups":  [],
            "treeLevel": 1
        }
        return item
    
    def __createScenarioItem(self, id, name, elapsed, isSkipped):
        if not isSkipped:
            content = f"<h4>{name}</h4><div><i>elapsed:{round(elapsed,2)}</i></div>"
            item = {
                "id" : id,
                "content": content
            }
            return item
        else:
            content = f"<h4>{name}</h4><div><i>elapsed:{round(elapsed,2)} (s)</i></div><div><i>Skipped</i></div>"
            item = {
                "id" : id,
                "content": content,
                "className": "skipped-scenario"
            }
            return item
    
    def __createScenarioBackgroundItem(self, id, startTime, endTime):
        item = {
            "id": id,
            "content": "",
            "group": id,
            "start": startTime.strftime("%m/%d/%Y, %H:%M:%S"),
            "end": endTime.strftime("%m/%d/%Y, %H:%M:%S"),
            "type": "background",
            "className": "negative",
        }
        return item

    def __createStepItem(self,itemId, groupId,step, scenarioStartTime):
        className = "default"
        if step["error"] is not None:
            className = "red"
        elif step["status"] == "skipped":
            className = "orange"
        item = {
            "id": itemId,
            "content": f"{step['keyword']}{step['text']}",
            "group": groupId,
            "start": step["start"].strftime("%m/%d/%Y, %H:%M:%S") if step["start"] is not None else scenarioStartTime.strftime("%m/%d/%Y, %H:%M:%S"),
            "end": step["end"].strftime("%m/%d/%Y, %H:%M:%S") if step["end"] is not None else scenarioStartTime.strftime("%m/%d/%Y, %H:%M:%S"),
            "type": "range",
            "title": step["elapsed"],
            "className": className
        }
        return item
    

    def generateTimeline(self, taskReport: Any):
        print(f"Generate timeline...")

        groups = []
        items = []
        allSteps = {}
        allScenarios = {}
        for key, group in groupby(sorted(taskReport,key=lambda x:x["feature"]), lambda x: x["feature"]):
            feature = self.__createFeatureItem(key)
            
            groups.append(feature)

            for t in group:
                feature["nestedGroups"].append(t["id"])
                scenarioResult: ScenarioResult = t["scenario"]
                if scenarioResult:
                    scenarioItem = self.__createScenarioItem(t["id"], t["name"], scenarioResult.elapsed, False)
                    groups.append(scenarioItem)

                    scenarioBackgroundItem = self.__createScenarioBackgroundItem(t["id"], scenarioResult.startTime, scenarioResult.endTime)

                    items.append(scenarioBackgroundItem)

                    for step in scenarioResult.steps:
                        itemId = uuid.uuid4().hex
                        stepItem = self.__createStepItem(itemId,t["id"], step, scenarioResult.startTime)
                        items.append(stepItem)
                        allSteps[itemId] = step
                            

                    dict_filter = lambda x, y: dict([ (i,x[i]) for i in x if i in set(y) ])
                    allScenarios[t["id"]] = dict_filter(vars(scenarioResult), ("elapsed","pid","threadId","startTime", "endTime",))
                else:
                    scenarioItem = self.__createScenarioItem(t["id"], t["name"], t["elapsed"], True)
                    groups.append(scenarioItem)
                
                
        pgroups,pitems = self.__getAllPids(taskReport)


        cssContent = self.getTemplatePropertyContent('vis-timeline-graph2d.min.css')
        jsContent = self.getTemplatePropertyContent('vis-timeline-graph2d.min.js')

        template = self.templateEnv.get_template("timeline.html")
        output = template.render(groups=json.dumps(groups),
        css=cssContent,
        js=jsContent,
        items=json.dumps(items),
        steps=json.dumps(allSteps, default=datetimeConverter),
        scenarios=json.dumps(allScenarios, default=datetimeConverter),
        plistGroups=json.dumps(pgroups, default=datetimeConverter),
        plistItems=json.dumps(pitems, default=datetimeConverter)
        )

        self.writeTemplateContent("timeline_output.html",output)



import time
import sys
import os
from conclave.task_runner import TaskRunner
from conclave.step import Step
from conclave.world import World
import multiprocessing

from conclave.task_runner import TaskRunner

if __name__ == '__main__':
    print(f"cpu count: {multiprocessing.cpu_count()}")
    tr = TaskRunner(debugMode=True)
    testResult = tr.run(["concurrent.feature"])

    print("\nprogram elapsed time :", testResult.elapsed)

    if not testResult.success:
        print(f"Test failed")
        os._exit(1)